@@ -1,62 +1,98 @@
 from novelai_api.NovelAIError import NovelAIError
+from novelai_api.FakeClientSession import FakeClientSession
 
 from aiohttp import ClientSession, ClientError
 from aiohttp.client_reqrep import ClientResponse
 from aiohttp.client import _RequestContextManager
 from aiohttp.http_exceptions import HttpProcessingError
 
+from requests import request as sync_request
+
 from typing import Union, Dict, Tuple, List, Iterable, Any, NoReturn, Optional, MethodDescriptorType
 
 #=== INTERNALS ===#
-def treat_response_object(rsp: ClientResponse, content: Any, status: int) -> Any:
-	# success
-	if rsp.status == status:
-		return content
-
-	# not success, but valid response
-	if type(content) is dict and "message" in content:	# NovelAI REST API error
-		raise NovelAIError(rsp.status, content["message"])
-	elif hasattr(rsp, "reason"):						# HTTPException error
-		raise NovelAIError(rsp.status, str(rsp.reason))
-	else:
-		raise NovelAIError(rsp.status, "Unknown error")
-
-def treat_response_bool(rsp: ClientResponse, content: Any, status: int) -> bool:
-	if 100 <= rsp.status < 400:
-		return rsp.status == status
-
-	return bool(treat_response_object(rsp, content, status))
-
-async def treat_response(rsp: ClientResponse) -> Any:
-	if rsp.content_type == "application/json":
-		return await rsp.json()
-	else:
-		return await rsp.text()
-
 #=== API ===#
 class Low_Level:
 	_parent: "NovelAI_API"
-	_session: ClientSession
+	_session: Union[ClientSession, FakeClientSession]
+	_is_async: bool
 
 	def __init__(self, parent: "NovelAI_API"):
+		self._is_async = parent._is_async
+
+		assert not self._is_async or isinstance(parent._session, ClientSession), "Session must be of class ClientSession for asynchronous operations"
+		assert self._is_async or isinstance(parent._session, Fake), "Session must be of class FakeClientSession for synchronous operations"
+
 		self._parent = parent
 		self._session = parent._session
 
+	def _treat_response_object(self, rsp: ClientResponse, content: Any, status: int) -> Any:
+		# success
+		if rsp.status == status:
+			return content
+
+		# not success, but valid response
+		if type(content) is dict and "message" in content:	# NovelAI REST API error
+			raise NovelAIError(rsp.status, content["message"])
+		elif hasattr(rsp, "reason"):						# HTTPException error
+			raise NovelAIError(rsp.status, str(rsp.reason))
+		else:
+			raise NovelAIError(rsp.status, "Unknown error")
+
+	def _treat_response_bool(self, rsp: ClientResponse, content: Any, status: int) -> bool:
+		if 100 <= rsp.status < 400:
+			return rsp.status == status
+
+		return bool(treat_response_object(rsp, content, status))
+
+	async def _treat_response(self, rsp: ClientResponse) -> Any:
+		if rsp.content_type == "application/json":
+			return (await rsp.json()) if self._is_async else rsp.json()
+		else:
+			return (await rsp.text()) if self._is_async else rsp.text
+
+	async def _request_async(self, method: str, url: str, data: Optional[Union[Dict[str, Any], str]] = None) -> Tuple[ClientResponse, Any]:
+		"""
+		:param url: Url of the request
+		:param method: Method of the request from ClientSession
+		:param data: Data to pass to the method if needed
+		"""	
+		if type(data) is dict:	# data transforms dict in str
+			async with self._session.request(method, url, json = data) as rsp:
+				return (rsp, await self._treat_response(rsp))
+		else:
+			async with self._session.request(method, url, data = data) as rsp:
+				return (rsp, await self._treat_response(rsp))
+
+	async def _request_sync(self, method: str, url: str, data: Optional[Union[Dict[str, Any], str]] = None) -> Tuple[ClientResponse, Any]:
+		"""
+		:param url: Url of the request
+		:param method: Method of the request from the request library
+		:param data: Data to pass to the method if needed
+		"""
+
+		timeout = self._session.timeout.total
+
+		if type(data) is dict:
+			with sync_request(method, url, json = data) as rsp:
+				return (rsp, await self._treat_response(rsp))
+		else:
+			with sync_request(method, url, data = data) as rsp:
+				return (rsp, await self._treat_response(rsp))
+
 	async def request(self, method: str, endpoint: str, data: Optional[Union[Dict[str, Any], str]] = None) -> Tuple[ClientResponse, Any]:
 		"""
-		:param url: Endpoint of the request
+		:param endpoint: Endpoint of the request
 		:param request_method: Method of the reqest from ClientSession
-		:param data: Data to pass to the request_method if needed
+		:param data: Data to pass to the method if needed
 		"""
 
 		url = f"{self._parent._BASE_ADDRESS}{endpoint}"
 
-		if type(data) is dict:	# data transforms dict in str
-			async with self._session.request(method, url, json = data) as rsp:
-				return (rsp, await treat_response(rsp))
+		if self._is_async:
+			return await self._request_async(method, url, data)
 		else:
-			async with self._session.request(method, url, data = data) as rsp:
-				return (rsp, await treat_response(rsp))
+			return await self._request_sync(method, url, data)
 
 	async def is_reachable(self) -> bool:
 		"""
@@ -65,7 +101,7 @@ class Low_Level:
 		:return: True if reachable, False if not
 		"""
 		rsp, content = await self.request("get", "/")
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def register(self, recapcha: str, access_key: str, email: Optional[str], giftkey: Optional[str] = None) -> bool:
 		"""
@@ -95,7 +131,7 @@ class Low_Level:
 			data["giftkey"] = giftkey
 
 		rsp, content = await self.request("post", "/user/register", data)
-		rsp = treat_response_bool(rsp, content, 201)
+		rsp = self._treat_response_bool(rsp, content, 201)
 
 		# FIXME: handle cases where the response is corrupted
 		self._parent._token = rsp["accessToken"]
@@ -108,7 +144,7 @@ class Low_Level:
 
 		:param access_key: Access key of the account
 
-		:return: True on success
+		:return: Response of the request
 		"""
 
 		assert type(access_key) is str, f"Expected type 'str' for access_key, but got type '{type(access_key)}'"
@@ -116,7 +152,7 @@ class Low_Level:
 		assert len(access_key) == 64, f"access_key should be 64 characters, got length of {len(access_key)}"
 
 		rsp, content = await self.request("post", "/user/login", { "key": access_key })
-		return treat_response_object(rsp, content, 201)
+		return self._treat_response_object(rsp, content, 201)
 
 	async def change_access_key(self, current_key: str, new_key: str) -> bool:
 		assert type(current_key) is str, f"Expected type 'str' for current_key, but got type '{type(current_key)}'"
@@ -126,13 +162,13 @@ class Low_Level:
 		assert len(new_key) == 64, f"New access key should be 64 characters, got length of {len(new_key)}"
 
 		rsp, content = await self.request("post", "/user/change-access-key", { "currentAccessKey": current_key, "newAccessKey": new_key })
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def request_account_recovery(self, email: str) -> bool:
 		assert type(email) is str, f"Expected type 'str' for email, but got type '{type(email)}'"
 
 		rsp, content = await self.request("post", "/user/recovery/request", { "email": email })
-		return treat_response_bool(rsp, content, 202)
+		return self._treat_response_bool(rsp, content, 202)
 
 	async def recover_account(self, recovery_token: str, new_key: str, delete_content: bool = False) -> bool:
 		assert type(recovery_token) is str, f"Expected type 'str' for recovery_token, but got type '{type(recovery_token)}'"
@@ -143,35 +179,35 @@ class Low_Level:
 		assert len(new_key) == 64, f"New access key should be 64 characters, got length of {len(new_key)}"
 
 		rsp, content = await self.request("post", "//user/recovery/recover", { "recoveryToken": recovery_token, "newAccessKey": new_key, "deleteContent": delete_content })
-		return treat_response_bool(rsp, content, 201)
+		return self._treat_response_bool(rsp, content, 201)
 
 	async def delete_account(self) -> bool:
 		rsp, content = await self.request("post", "/user/delete", None)
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def get_priority(self) -> Dict[str, Any]:
 		rsp, content = await self.request("get", "/user/priority")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def get_subscription(self) -> Dict[str, Any]:
 		rsp, content = await self.request("get", "/user/subscription")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def get_keystore(self) -> Dict[str, str]:
 		rsp, content = await self.request("get", "/user/keystore")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def set_keystore(self, keystore: Dict[str, str]) -> bool:
 		assert type(keystore) is dict, f"Expected type 'dicy' for keystore, but got type '{type(keystore)}'"
 
 		rsp, content = await self.request("put", "/user/keystore", keystore)
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def download_objects(self, object_type: str) -> Dict[str, List[Dict[str, Union[str, int]]]]:
 		assert type(object_type) is str, f"Expected type 'str' for object_type, but got type '{type(object_type)}'"
 
 		rsp, content = await self.request("get", f"/user/objects/{object_type}")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def upload_objects(self, object_type: str, meta: str, data: str) -> bool:
 		assert type(object_type) is str, f"Expected type 'str' for object_type, but got type '{type(object_type)}'"
@@ -181,14 +217,14 @@ class Low_Level:
 		assert len(meta) <= 128, f"Meta should be at most 128 characters, got length of {len(meta)}"
 
 		rsp, content = await self.request("put", f"/user/objects/{object_type}", { "meta": meta, "data": data })
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def download_object(self, object_type: str, object_id: str) -> Dict[str, Union[str, int]]:
 		assert type(object_type) is str, f"Expected type 'str' for object_type, but got type '{type(object_type)}'"
 		assert type(object_id) is str, f"Expected type 'str' for object_id, but got type '{type(object_id)}'"
 
 		rsp, content = await self.request("get", f"/user/objects/{object_type}/{object_id}")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def upload_object(self, object_type: str, object_id: str, meta: str, data: str) -> bool:
 		assert type(object_type) is str, f"Expected type 'str' for object_type, but got type '{type(object_type)}'"
@@ -199,37 +235,37 @@ class Low_Level:
 		assert len(meta) <= 128, f"Meta should be at most 128 characters, got length of {len(meta)}"
 
 		rsp, content = await self.request("patch", f"/user/objects/{object_type}/{object_id}", { "meta": meta, "data": data })
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def delete_object(self, object_type: str, object_id: str) -> Dict[str, Union[str, int]]:
 		assert type(object_type) is str, f"Expected type 'str' for object_type, but got type '{type(object_type)}'"
 		assert type(object_id) is str, f"Expected type 'str' for object_id, but got type '{type(object_id)}'"
 
 		rsp, content = await self.request("delete", f"/user/objects/{object_type}/{object_id}")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def get_settings(self) -> str:
 		rsp, content = await self.request("get", "/user/clientsettings")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def set_settings(self, value: str) -> bool:
 		assert type(value) is str, f"Expected type 'str' for value, but got type '{type(value)}'"
 
 		rsp, content = await self.request("put", "/user​/clientsettings", value)
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def bind_subscription(self, payment_processor: str, subscription_id: str) -> bool:
 		assert type(payment_processor) is str, f"Expected type 'str' for payment_processor, but got type '{type(payment_processor)}'"
 		assert type(subscription_id) is str, f"Expected type 'str' for subscription_id, but got type '{type(subscription_id)}'"
 
 		rsp, content = await self.request("post", "/user/subscription/bind", { "paymentProcessor": payment_processor, "subscriptionId": subscription_id })
-		return treat_response_bool(rsp, content, 201)
+		return self._treat_response_bool(rsp, content, 201)
 
 	async def change_subscription(self, new_plan: str) -> bool:
 		assert type(new_plan) is str, f"Expected type 'str' for new_plan, but got type '{type(new_plan)}'"
 
 		rsp, content = await self.request("post", "​/user​/subscription​/change", { "newSubscriptionPlan": new_plan })
-		return treat_response_bool(rsp, content, 200)
+		return self._treat_response_bool(rsp, content, 200)
 
 	async def generate(self, input: str, model: str, params: Dict[str, Any]) -> Dict[str, str]:
 		"""
@@ -250,7 +286,7 @@ class Low_Level:
 		args = { "input": input, "model": model, "parameters": params }
 
 		rsp, content = await self.request("post", "/ai/generate", args)
-		return treat_response_object(rsp, content, 201)
+		return self._treat_response_object(rsp, content, 201)
 
 	async def generate_stream(self):
 		raise NotImplementedError("Function is not implemented yet")
@@ -276,7 +312,7 @@ class Low_Level:
 		assert type(desc) is str, f"Expected type 'str' for desc, but got type '{type(desc)}'"
 
 		rsp, content = await self.request("post", "​/ai​/module​/train", { "data": data, "lr": rate, "steps": steps, "name": name, "description": desc })
-		return treat_response_object(rsp, content, 201)
+		return self._treat_response_object(rsp, content, 201)
 
 	async def get_modules(self) -> List[Dict[str, Any]]:
 		"""
@@ -284,7 +320,7 @@ class Low_Level:
 		"""
 
 		rsp, content = await self.request("get", "/ai/module/all")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def get_module(self, module_id: str) -> Dict[str, Any]:
 		"""
@@ -296,7 +332,7 @@ class Low_Level:
 		assert type(module_id) is str, f"Expected type 'str' for module_id, but got type '{type(module_id)}'"
 
 		rsp, content = await self.request("get", f"/ai/module/{module_id}")
-		return treat_response_object(rsp, content, 200)
+		return self._treat_response_object(rsp, content, 200)
 
 	async def delete_module(self, module_id: str) -> Dict[str, Any]:
 		"""
@@ -310,4 +346,4 @@ class Low_Level:
 		assert type(module_id) is str, f"Expected type 'str' for module_id, but got type '{type(module_id)}'"
 
 		rsp, content = await self.request("delete", f"/ai/module/{module_id}")
-		return treat_response_object(rsp, content, 200)
\ No newline at end of file
+		return self._treat_response_object(rsp, content, 200)
\ No newline at end of file
